import numpy as np
import cv2
import math
import copy

def read_transparent_png(filename):
    image_4channel = cv2.imread(filename, cv2.IMREAD_UNCHANGED)
    alpha_channel = image_4channel[:,:,3]
    rgb_channels = image_4channel[:,:,:3]

    # White Background Image
    white_background_image = np.ones_like(rgb_channels, dtype=np.uint8) * 255

    # Alpha factor
    alpha_factor = alpha_channel[:,:,np.newaxis].astype(np.float32) / 255.0
    alpha_factor = np.concatenate((alpha_factor,alpha_factor,alpha_factor), axis=2)

    # Transparent Image Rendered on White Background
    base = rgb_channels.astype(np.float32) * alpha_factor
    white = white_background_image.astype(np.float32) * (1 - alpha_factor)
    final_image = base + white
    return final_image.astype(np.uint8)

def padRightDownCorner(img, stride, padValue):
    h = img.shape[0]
    w = img.shape[1]

    pad = 4 * [None]
    pad[0] = 0  # up
    pad[1] = 0  # left
    pad[2] = 0 if (h % stride == 0) else stride - (h % stride)  # down
    pad[3] = 0 if (w % stride == 0) else stride - (w % stride)  # right

    img_padded = img
    pad_up = np.tile(img_padded[0:1, :, :] * 0 + padValue, (pad[0], 1, 1))
    img_padded = np.concatenate((pad_up, img_padded), axis=0)
    pad_left = np.tile(img_padded[:, 0:1, :] * 0 + padValue, (1, pad[1], 1))
    img_padded = np.concatenate((pad_left, img_padded), axis=1)
    pad_down = np.tile(img_padded[-2:-1, :, :] * 0 + padValue, (pad[2], 1, 1))
    img_padded = np.concatenate((img_padded, pad_down), axis=0)
    pad_right = np.tile(img_padded[:, -2:-1, :] * 0 + padValue, (1, pad[3], 1))
    img_padded = np.concatenate((img_padded, pad_right), axis=1)

    return img_padded, pad


# transfer caffe model to pytorch which will match the layer name
def transfer(model, model_weights):
    transfered_model_weights = {}
    for weights_name in model.state_dict().keys():
        transfered_model_weights[weights_name] = model_weights['.'.join(weights_name.split('.')[1:])]
    return transfered_model_weights


def dumpRotateImage(img, degree):
    height, width = img.shape[:2]
    heightNew = int(
        width * math.fabs(math.sin(math.radians(degree))) + height * math.fabs(math.cos(math.radians(degree))))
    widthNew = int(
        height * math.fabs(math.sin(math.radians(degree))) + width * math.fabs(math.cos(math.radians(degree))))
    matRotation = cv2.getRotationMatrix2D((width // 2, height // 2), degree, 1)
    matRotation[0, 2] += (widthNew - width) // 2
    matRotation[1, 2] += (heightNew - height) // 2
    # imgRotation = cv2.warpAffine(img, matRotation, (widthNew, heightNew), borderValue=(255, 255, 255))
    imgRotation = cv2.warpAffine(img, matRotation, (widthNew, heightNew), borderValue=(0, 0, 0))
    return imgRotation, matRotation


def cvt_coor(x_global, y_global, x_local, y_local, matRotation):
    pt1 = np.dot(matRotation, np.array([[x_local], [y_local], [1]]))
    return int((x_global + pt1[0])[0]), int((y_global + pt1[1])[0])


def draw_body_part(i, canvas, body_img, x, y, angle, scale, scale_factor=1):
    body_img_h, body_img_w, _ = body_img.shape
    body_img = cv2.resize(body_img, (int(body_img_w * scale * scale_factor), int(body_img_h * scale * scale_factor)))
    body_img_h, body_img_w, _ = body_img.shape
    x1_body_img, y1_body_img, x2_body_img, y2_body_img = int(body_img_w / 2), 0, int(body_img_w / 2), body_img_h - 1
    imgRotation, matRotation = dumpRotateImage(body_img, angle)
    imgRotation[imgRotation == 255] = 0
    if not i:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[0] - pt1[0])[0]), int((y[0] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, 0, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
        x_y = cvt_coor(x_global, y_global, 0, int(body_img_h / 3), matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
        x_y = cvt_coor(x_global, y_global, body_img_w - 1, int(body_img_h / 3), matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
        x_y = cvt_coor(x_global, y_global, int(body_img_w / 5), body_img_h - 1, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
        x_y = cvt_coor(x_global, y_global, int(body_img_w / 5 * 4), body_img_h - 1, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 1:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y2_body_img], [1]]))
        x_global, y_global = int((x[1] - pt1[0])[0]), int((y[1] - pt1[1])[0])
    elif i == 2:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[2] - pt1[0])[0]), int((y[2] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 3:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[6] - pt1[0])[0]), int((y[6] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 4:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[3] - pt1[0])[0]), int((y[3] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 5:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[8] - pt1[0])[0]), int((y[8] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 6:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[4] - pt1[0])[0]), int((y[4] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 7:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[10] - pt1[0])[0]), int((y[10] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 8:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[5] - pt1[0])[0]), int((y[5] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    elif i == 9:
        pt1 = np.dot(matRotation, np.array([[x1_body_img], [y1_body_img], [1]]))
        x_global, y_global = int((x[12] - pt1[0])[0]), int((y[12] - pt1[1])[0])
        x_y = cvt_coor(x_global, y_global, x1_body_img, y2_body_img, matRotation)
        x.append(x_y[0])
        y.append(x_y[1])
    try:
        if (y_global + imgRotation.shape[0]) < canvas.shape[0] and (x_global + imgRotation.shape[1]) < canvas.shape[1]:
            canvas[y_global:y_global + imgRotation.shape[0], x_global:x_global + imgRotation.shape[1],
            :] = cv2.bitwise_or(
                canvas[y_global:y_global + imgRotation.shape[0], x_global:x_global + imgRotation.shape[1], :],
                imgRotation)
    except:
        pass

    return canvas, x, y


# draw the body keypoint and lims
def draw_anim(canvas, body_img_list, candidate, subset):
    limbSeq = [[1, 8, 11], [0, 1], [2, 3], [3, 4], [5, 6], [6, 7], [8, 9], [9, 10], [11, 12], [12, 13]]
    for n in range(len(subset)):
        # TODO add scale refresh
        scale = None
        x, y = [], []
        for i in range(len(limbSeq)):
            index = subset[n][np.array(limbSeq[i])]
            if -1 in index:
                continue
            X = candidate[index.astype(int), 0]
            Y = candidate[index.astype(int), 1]
            X = list(map(int, X))
            Y = list(map(int, Y))
            if len(X) == 3 and len(Y) == 3:
                X = [X[0], int(X[1] / 2 + X[2] / 2)]
                Y = [Y[0], int(Y[1] / 2 + Y[2] / 2)]
            if not i:
                x.append(X[0])
                y.append(Y[0])
            body_img = copy.deepcopy(body_img_list[i])
            body_img[body_img == 255] -= 1

            angle = (math.degrees(math.atan2(Y[0] - Y[1], X[0] - X[1])) + 90) * -1
            if not scale:
                length = ((X[0] - X[1]) ** 2 + (Y[0] - Y[1]) ** 2) ** 0.5
                body_img_h, body_img_w, _ = body_img.shape
                scale = length / body_img_w
            try:
                canvas, x, y = draw_body_part(i, canvas, body_img, x, y, angle, scale, scale_factor=1)
            except:
                continue

    return canvas
